<?php

namespace Oro\Bundle\ShippingBundle\Tests\Unit\Form\Extension;

use Doctrine\Common\Persistence\ManagerRegistry;
use Doctrine\Common\Persistence\ObjectManager;
use Doctrine\Common\Persistence\ObjectRepository;
use Oro\Bundle\ProductBundle\Entity\Product;
use Oro\Bundle\ProductBundle\Form\Type\ProductType;
use Oro\Bundle\ShippingBundle\Form\Extension\ProductFormExtension;
use Oro\Bundle\ShippingBundle\Form\Type\ProductShippingOptionsCollectionType;
use Oro\Bundle\ShippingBundle\Validator\Constraints\UniqueProductUnitShippingOptions;
use Oro\Component\Testing\Unit\EntityTrait;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Form\FormEvent;
use Symfony\Component\Form\FormEvents;
use Symfony\Component\Form\FormInterface;

/**
 * @SuppressWarnings(PHPMD.TooManyMethods)
 */
class ProductFormExtensionTest extends \PHPUnit\Framework\TestCase
{
    use EntityTrait;

    /** @var ProductFormExtension */
    protected $extension;

    /** @var ObjectManager|\PHPUnit\Framework\MockObject\MockObject */
    protected $manager;

    /** @var ObjectRepository|\PHPUnit\Framework\MockObject\MockObject */
    protected $repo;

    /** @var ManagerRegistry|\PHPUnit\Framework\MockObject\MockObject */
    protected $registry;

    protected function setUp(): void
    {
        $this->repo = $this->getMockBuilder('Doctrine\Common\Persistence\ObjectRepository')
            ->disableOriginalConstructor()
            ->getMock();

        $this->manager = $this->createMock('Doctrine\Common\Persistence\ObjectManager');
        $this->manager->expects($this->any())
            ->method('getRepository')
            ->with('OroShippingBundle:ProductShippingOptions')
            ->willReturn($this->repo);

        $this->registry = $this->createMock('Doctrine\Common\Persistence\ManagerRegistry');
        $this->registry->expects($this->any())
            ->method('getManagerForClass')
            ->with('OroShippingBundle:ProductShippingOptions')
            ->willReturn($this->manager);

        $this->extension = new ProductFormExtension($this->registry);
    }

    /**
     * @param mixed $product
     *
     * @dataProvider formDataProvider
     */
    public function testBuildForm($product)
    {
        /** @var FormBuilderInterface|\PHPUnit\Framework\MockObject\MockObject $builder */
        $builder = $this->createMock('Symfony\Component\Form\FormBuilderInterface');
        $builder->expects($this->once())
            ->method('add')
            ->with(
                ProductFormExtension::FORM_ELEMENT_NAME,
                ProductShippingOptionsCollectionType::class,
                [
                    'label' => 'oro.shipping.product_shipping_options.entity_plural_label',
                    'required' => false,
                    'mapped' => false,
                    'constraints' => [new UniqueProductUnitShippingOptions()],
                    'entry_options' => [
                        'product' => $product,
                    ],
                ]
            );

        $builder->expects($this->exactly(3))->method('addEventListener');
        $builder->expects($this->at(2))
            ->method('addEventListener')
            ->with(FormEvents::POST_SET_DATA, [$this->extension, 'onPostSetData']);
        $builder->expects($this->at(3))
            ->method('addEventListener')
            ->with(FormEvents::PRE_SUBMIT, [$this->extension, 'onPreSubmit']);
        $builder->expects($this->at(4))
            ->method('addEventListener')
            ->with(FormEvents::POST_SUBMIT, [$this->extension, 'onPostSubmit'], 10);

        $builder->expects($this->once())->method('getData')->willReturn($product);

        $this->extension->buildForm($builder, []);
    }

    public function testGetExtendedTypes()
    {
        $this->assertEquals([ProductType::class], ProductFormExtension::getExtendedTypes());
    }

    /**
     * @dataProvider formDataProvider
     *
     * @param null|Product $product
     */
    public function testOnPostSetData($product)
    {
        $event = $this->createEvent($product);

        /** @var FormInterface|\PHPUnit\Framework\MockObject\MockObject $form */
        $form = $event->getForm()->get(ProductFormExtension::FORM_ELEMENT_NAME);

        $form->expects($product ? $this->once() : $this->never())->method('setData');

        $this->extension->onPostSetData($event);
    }

    public function testOnPostSubmitEmptyProduct()
    {
        $event = $this->createEvent();
        /** @var FormInterface|\PHPUnit\Framework\MockObject\MockObject $form */
        $form = $event->getForm()->get(ProductFormExtension::FORM_ELEMENT_NAME);

        $form->expects($this->never())->method('getData');

        $this->extension->onPostSubmit($event);
    }

    public function testOnPostSubmitValidData()
    {
        $product = $this->createMockProduct();

        $event = $this->createEvent($this->createMockProduct());
        /** @var FormInterface|\PHPUnit\Framework\MockObject\MockObject $form */
        $form = $event->getForm()->get(ProductFormExtension::FORM_ELEMENT_NAME);

        $event->getForm()->expects($this->once())->method('isValid')->willReturn(true);

        $removedOption = $this->getEntity('Oro\Bundle\ShippingBundle\Entity\ProductShippingOptions', ['id' => 42]);

        $this->repo->expects($this->once())
            ->method('findBy')
            ->with(['product' => $product], ['productUnit' => 'ASC'])
            ->willReturn([$removedOption]);

        $form->expects($this->once())
            ->method('getData')
            ->willReturn(
                [
                    $this->getEntity('Oro\Bundle\ShippingBundle\Entity\ProductShippingOptions', ['id' => 1]),
                    $this->getEntity('Oro\Bundle\ShippingBundle\Entity\ProductShippingOptions', ['id' => null]),
                ]
            );

        $this->manager->expects($this->exactly(2))->method('persist');
        $this->manager->expects($this->once())->method('remove')->with($removedOption);

        $this->extension->onPostSubmit($event);
    }

    public function testOnPostSubmitInvalidData()
    {
        $event = $this->createEvent($this->createMockProduct());
        /** @var FormInterface|\PHPUnit\Framework\MockObject\MockObject $form */
        $form = $event->getForm()->get(ProductFormExtension::FORM_ELEMENT_NAME);

        $event->getForm()->expects($this->once())->method('isValid')->willReturn(false);

        $form->expects($this->once())
            ->method('getData')
            ->willReturn([$this->getEntity('Oro\Bundle\ShippingBundle\Entity\ProductShippingOptions', ['id' => 1])]);

        $this->manager->expects($this->never())->method('persist');

        $this->extension->onPostSubmit($event);
    }

    public function testOnPreSubmit()
    {
        $product = $this->createMockProduct();

        $event = $this->createEvent(
            [
                ProductFormExtension::FORM_ELEMENT_NAME => [
                    5 => ['productUnit' => 'test2'],
                    10 => ['productUnit' => 'test1']
                ]
            ]
        );
        $event->expects($this->once())
            ->method('setData')
            ->with(
                [
                    ProductFormExtension::FORM_ELEMENT_NAME => [
                        0 => ['productUnit' => 'test1'],
                        5 => ['productUnit' => 'test2']
                    ]
                ]
            );

        /** @var FormInterface|\PHPUnit\Framework\MockObject\MockObject $form */
        $form = $event->getForm();
        $form->expects($this->once())->method('getData')->willReturn($product);

        /** @var FormInterface|\PHPUnit\Framework\MockObject\MockObject $childForm */
        $childForm = $form->get(ProductFormExtension::FORM_ELEMENT_NAME);
        $childForm->expects($this->once())
            ->method('getData')
            ->willReturn(
                [
                    $this->getEntity(
                        'Oro\Bundle\ShippingBundle\Entity\ProductShippingOptions',
                        [
                            'id' => 42,
                            'productUnit' => $this->getEntity(
                                'Oro\Bundle\ProductBundle\Entity\ProductUnit',
                                ['code' => 'test1']
                            )
                        ]
                    )
                ]
            );

        $this->extension->onPreSubmit($event);
    }

    public function testOnPreSubmitWithoutProduct()
    {
        $product = $this->createMockProduct(null);

        $event = $this->createEvent();
        $event->expects($this->never())->method('setData');

        /** @var FormInterface|\PHPUnit\Framework\MockObject\MockObject $form */
        $form = $event->getForm();
        $form->expects($this->once())->method('getData')->willReturn($product);

        $this->extension->onPreSubmit($event);
    }

    /**
     * @param object|null $data
     *
     * @return FormEvent|\PHPUnit\Framework\MockObject\MockObject
     */
    protected function createEvent($data = null)
    {
        $form = $this->createMock('Symfony\Component\Form\FormInterface');

        /** @var FormInterface|\PHPUnit\Framework\MockObject\MockObject $mainForm */
        $mainForm = $this->createMock('Symfony\Component\Form\FormInterface');
        $mainForm->expects($this->any())
            ->method('get')
            ->with(ProductFormExtension::FORM_ELEMENT_NAME)
            ->willReturn($form);

        /** @var FormEvent|\PHPUnit\Framework\MockObject\MockObject $event */
        $event = $this->getMockBuilder('Symfony\Component\Form\FormEvent')->disableOriginalConstructor()->getMock();
        $event->expects($this->any())->method('getForm')->willReturn($mainForm);
        $event->expects($this->any())->method('getData')->willReturn($data);

        return $event;
    }

    /**
     * @return array
     */
    public function formDataProvider()
    {
        $productMock = $this->createMockProduct();

        return [
            'no product' => ['product' => null],
            'with product' => ['product' => $productMock],
        ];
    }

    /**
     * @param int $id
     * @return Product
     */
    private function createMockProduct($id = 1)
    {
        return $this->getEntity('Oro\Bundle\ProductBundle\Entity\Product', ['id' => $id]);
    }
}
